import sys
import os

rapid_api_token = "bc5b270c17msh058aaac04af67a4p16d4ccjsn75cd807f68f5"
google_api_token = "AIzaSyB94FUFckhQUa-jRpy8Us8HF7AeRQ5KPqU"
estated_token = 'J9aEmitNylIyFD5bQiVLdsZs07qw1t'
mongo_pass = 'Barca2004!'
