"""test the PeekGeo package"""
"""from placesGeocoding.peek_geo import peekgeo_api as pg
from time import perf_counter

# create geocoder object
geocoder = pg.PeekGeo('eyJhbGci')

start = perf_counter()
# compute query output
print(geocoder.batch_forward(
    ['20 Manor House Lane, Dobbs Ferry, NY, 10522',
     '18 Manor House Lane, Dobbs Ferry, NY, 10522',
     '16 Manor House Lane, Dobbs Ferry, NY, 10522',
     '29 Manor House Lane, Dobbs Ferry, NY, 10522',
     '31 Manor House Lane, Dobbs Ferry, NY, 10522',
     '35 Manor House Lane, Dobbs Ferry, NY, 10522']
))
print(perf_counter() - start)"""

