"""address autocomplete/correction and address parsing models go here"""

from pydantic import BaseModel
from typing import List, Union, Tuple, Dict


class CorrectionModel(BaseModel):
    """dummy correction model"""
    pass